#ifndef _cxxtls_Exception_header_included_p
#define _cxxtls_Exception_header_included_p

//
// Copyright 2015, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure are covered by the
// same copyright.  Some of them are part of the GNU source distribution
// and you must obey the GPL copyright for those files.

#include <cxxtls/lexical_cast.h>
#include <string>
#include <exception>

///@file
///
///  This file defines class cxxtls::Exception.  This class derives from
///  std::exception but has constructors which make it easy to construct
///  error message style exceptions.  
///
///  Also declare the macro, DECLARE_CXXTLS_EXCEPTION, which is used to build
///  exceptions.
///
///  Also, DECLARE_CXXTLS_PARENTED_EXCEPTION, which does the same thing but
///  also makes the declared exception be derived from a user specfied base
///  class as well as cxxtls::exception.  This lets you use dynamic casts on your
///  exception to check for special cases.

namespace cxxtls
{

    class Exception
    :  public std::exception
        ///
        ///   You should probably be using either one of the following macros:
        ///
        ///   @li DECLARE_CXXTLS_EXCEPTION 
        ///   @li DECLARE_CXXTLS_PARENTED_EXCEPTION.
        ///
        ///   Class Exception derives from std::exception and adds the following features:
        ///
        ///   It assumes that there is a std::string that implements the exception body
        ///   and it provides multiple templatized constructors which allow for easier population of
        ///   the string.  You can also print the Exception to a ostream directly without calling what.
        ///
        ///   Consider the following constructor examples:
        ///
        ///   <table>
        ///   <tr><td> Exception()                       </td><td>"cxxtls::Exception"   </td></tr>
        ///   <tr><td> Exception("type")                 </td><td>"type"                </td></tr>
        ///   <tr><td> Exception("type, "text", 9)       </td><td>"type text=9"         </td></tr>
        ///   <tr><td> Exception("type", "a", 9, "b",8)  </td><td>"type a=9 b=8"        </td></tr>
        ///   </table>
        ///
        ///   In general, you construct an Exception with the type of the exception as a char const pointer,
        ///   and you can then specify up to 3 pairs of secondary parameters.  The secondary parameters
        ///   as follows:
        ///   @li  the name of the parameter to display to in the what() string
        ///   @li  a number or string to display as the value of the parameter.
        ///
        ///   cxxtls::lexical_cast is used to convert the values into a string, so you can actually
        ///   types other than number for the parameter values, but you may need to provide your own
        ///   implmentation of the lexical_cast<string> for your data types.
        ///
    {
        std::string message_;

    public:

        virtual char const *what() const throw() { return message_.c_str(); }   // handle the inherited member

        Exception()
        :  message_("cxxtls::exception")
        {
           
        }

        virtual ~Exception() throw ()
        {
        }


        template<class Ostream>
        friend Ostream &operator<< (Ostream &s, Exception const &me)
        {
           return s << me.message_;
        }


        Exception(char const *type)
        :  message_(type)
        {
        }

        template<class T1>
        Exception(char const *type, char const *t1_name, T1 const &v1)
        :  message_(type)
        {
           message_ += ' ';
           message_ += t1_name;
           message_ += '=';
           message_ += cxxtls::lexical_cast<std::string>(v1);
        }


        template<class T1, class T2>
        Exception(char const *type, char const *t1_name, T1 const &v1, char const *t2_name, T2 const &v2)
        :  message_(type)
        {
           message_ += ' ';
           message_ += t1_name;
           message_ += '=';
           message_ += cxxtls::lexical_cast<std::string>(v1);

           message_ += ',';
           message_ += t2_name;
           message_ += '=';
           message_ += cxxtls::lexical_cast<std::string>(v2);



        }

        template<class T1, class T2, class T3>
        Exception(char const *type, char const *t1_name, T1 const &v1, char const *t2_name, T2 const &v2, char const *t3_name, T3 const &v3)
        :  message_(type)
        {
           message_ += ' ';
           message_ += t1_name;
           message_ += '=';
           message_ += cxxtls::lexical_cast<std::string>(v1);

           message_ += ',';
           message_ += t2_name;
           message_ += '=';
           message_ += cxxtls::lexical_cast<std::string>(v2);


           message_ += ',';
           message_ += t3_name;
           message_ += '=';
           message_ += cxxtls::lexical_cast<std::string>(v3);

        }

    };  // class Exception


}  // namespace cxxtls





/// Macro DECLARE_BRASH_EXCEPTION is used to declare exception classes.
///  The exceptions so
///  created are derived from std::exception and they all have the
///  ability to accept parameters to the exception constructor that
///  let you associate values with the exception.  The values you add to
///  the constructor call are converted to strings and available as part
///  of the what() call's return value.  Note that when passing parameters
///  to the constructor, up to 3 pairs of parameters can be supplied.  Each
///  parameter pair consists of a "name" and  "value".  All names must be
///  a c_style char pointer string.  The value can be any type that works
///  with cxxtls::lexical_cast<>.  Here is an example:
///
///@begincode
///
///   throw SomeException("fileHandle", fileno_, "fileName", filename_, "sizePercent", 4.9);
///
///@endcode
///
///   When handling the exception, the what() function would return:  
///   "SomeException: fileHandle='456' fileName='someName' sizePercent='4.9'"





#define DECLARE_CXXTLS_EXCEPTION(name)   \
    struct name \
    : public cxxtls::Exception \
    { \
      name() \
      : cxxtls::Exception(#name) \
      {\
      }\
      template<class T1> \
      name(char const *t1_name, T1 const &v1) \
      :  cxxtls::Exception(#name, t1_name, v1) \
      {\
      }\
      template<class T1, class T2> \
      name(char const *t1_name, T1 const &v1, char const *t2_name, T2 const &v2) \
      :  cxxtls::Exception(#name, t1_name, v1, t2_name, v2) \
      {\
      }\
      template<class T1, class T2, class T3> \
      name(char const *t1_name, T1 const &v1, char const *t2_name, T2 const &v2, const char *t3_name, T3 const &v3) \
      :  cxxtls::Exception(#name, t1_name, v1, t2_name, v2, t3_name, v3) \
      {\
      }\
    }

/// DECLARE_CXXTLS_PARENTED_EXCEPTION is a macro that declares an exception, as above, but with a specified @b additional base class.
/// It also makes the exceptions derive from std::exception.  Use dynamic cast to figure out which
/// of the derived classes it actually is at run time.  For example:
///
///@begincode
///
///     struct MyBase {};
///
///     DECLARE_PARENTED_EXCEPTION(MyException, MyBase);
///
///     ...
///
///     catch( std::exception &e)
///     {
///         if(dynamic_cast<MyBase*>(&e))
///         {
///              ... // handle special cases
///         }
///
///         // handle normal exceptions
///     }
///@endcode
///
#define DECLARE_CXXTLS_PARENTED_EXCEPTION(name, parent)   \
    struct name \
    : public cxxtls::Exception \
    , parent \
    { \
      name() \
      : cxxtls::Exception(#name) \
      {\
      }\
      template<class T1> \
      name(char const *t1_name, T1 const &v1) \
      :  cxxtls::Exception(#name, t1_name, v1) \
      {\
      }\
      template<class T1, class T2> \
      name(char const *t1_name, T1 const &v1, char const *t2_name, T2 const &v2) \
      :  cxxtls::Exception(#name, t1_name, v1, t2_name, v2) \
      {\
      }\
      template<class T1, class T2, class T3> \
      name(char const *t1_name, T1 const &v1, char const *t2_name, T2 const &v2, const char *t3_name, T3 const &v3) \
      :  cxxtls::Exception(#name, t1_name, v1, t2_name, v2, t3_name, v3) \
      {\
      }\
    }



#endif
