#ifndef gdbcontroler_h_included_p
#define gdbcontroler_h_included_p

#ifdef __linux__
//
// NOT currently for windows
//

//
// Copyright 2013, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

#include <cxxtls/foreach.h>
#include <cxxtls/thread.h>
#include <cxxtls/controlledprocess.h>
#include <cxxtls/pseudoterminal.h>
#include <cxxtls/varobj.h>

#include <list>
#include <map>
#include <set>
#include <sys/types.h>
#include <portable_io.h>
#include <memory>
#include <vector>

///@file
///  This file defines GdbController, a C++ interface to gdb.
///
///  See http://sourceware.org/gdb/onlinedocs/gdb/GDB_002fMI.html#GDB_002fMI

namespace cxxtls
{
void* gdbControllerReader(void *p); // The thread that reads gdb output.  
                                    // Pass &GdbController
                                    // on thread create.

void* gdbControllerTgtReader(void *); // The thread that reads gdb output.  
                                      // Pass &GdbController
                                      // on thread create.

class GdbController
    /// GdbController is a class that encapsulates the signficant complexities
    /// involved in automatically interacting with GDB as it allows the 
    /// debugging of a child process.   It provides a fairly clean and simple
    /// interface that allows you to access the state of the debugged child
    /// as well as set breakpoints, list information, etc.
    ///
    /// Refer to the following gdb documentation to understand the programatic
    /// interface that is used by this class:
    ///
    /// @li http://sourceware.org/gdb/onlinedocs/gdb/GDB_002fMI.html#GDB_002fMI
    ///
    /// This class constructs threads to interact with GDB as it runs the
    /// child process.  So you must compile and link with -pthread.   Most
    /// class members are thread safe -- if not all.
    ///
    /// This class enforces the basic logic of 
    ///
    ///   @li starting gdb with the childprocess stopped
    ///   @li running the child process
    ///   @li waiting while the child does its business
    ///   @li interrupting the running process
    ///   @li killing the running child process
    ///
    /// To do so, GdbController issues commands into the gdb command stream and
    /// watches for results coming out of gdb and the debugged target process
    /// as it runs. 
    ///
    /// GDB is executed in a ControlledProcess.  When it runs the target process,
    /// ie. the child being debugged, it forces the child into a PseudoTerminal.
    /// This allows GdbController to know precisely which output is coming from 
    /// GDB itself and which is coming from the child.
    ///
    /// User input must be written to the PseudoTerminal's masterFD() so that it can be
    /// read from by the debugged child process.  The application using the GdbController
    /// can call the method sendToTarget() to write strings to the debugged child
    /// target process.
    ///
    /// Commands to GDB itself typically are encapsulated into methods like:
    ///
    ///   @li setArgs()
    ///   @li setExecutable()
    ///   @li run()
    ///   @li halt()
    ///   @li kill()
    ///   @li addBreakpoint()
    ///   @li addDisplayVariable()
    ///   @li breakpoints() -- get a list of the breakpoints.
    ///   @li writeTargetStdin() -- write a string to the target process' stdin stream.
    ///
    /// This class provides many interfaces.  See below.  Note that not
    /// all of the interfaces can be invoked without stopping the running
    /// process.  If they are used when the process is running, it will be
    /// temporarily halted to gather then information, then the process will
    /// be continued.  This can have undesireable effects on the program
    /// so be advised.  
    ///
    /// This interface attempts to match the design of GDB rather
    /// enforce some design of its own onto GDB.
    ///
    /// For an application to monitor the changes to the started/stopped
    /// state of a child process being debugged, it can read from the "stopped queue".
    /// See member waitForStop().
{
    GdbController &operator=(GdbController const &);   // suppress copy of any form
    GdbController(GdbController const&);

    friend void* gdbControllerReader(void *p);
    friend void* gdbControllerTgtReader(void *); 



public:
    enum ChildState { 
                        // keep this in sync with the array of name strings in gdbcontroller.cxx

                        notStarted,             ///<   the child has never been run
                        runnable,               ///<   ready to run but not yet run
                        running,                ///<   the child is currently running
                        stopped,                ///<   the child has received SIGINT
                        terminated,             ///<   the child has finished or been killed
                        unstartable,            ///<   gdb won't start or other major problem
                        gdbExiting,             ///<   gdb is shutting down

                        STATE_COUNT             // MUST BE LAST
                    };

    typedef ThreadQueue<char>  TargetOutput;    ///<   The type of queue holding the output from the debugged
                                                ///    child process.

    struct SignalHandling
        ///  A description of a signal and how it should be handled.
    {
       bool stop_;                              ///<  stop when the signal occurs
       bool print_;                             ///<  print and continue
       bool pass_;                              ///<  eat or pass the signal to the target
       std::string description_;                ///<  a description of the signal

       SignalHandling()
       :  stop_(false)
       ,  print_(false)
       ,  pass_(false)
       ,  description_("unknown")
       {
       }
    };

    typedef std::map<std::string     /*name*/,
                     SignalHandling  /*disposition*/
                    >
                    SignalHandlers;              ///<   A mapping between signal names
                                                 ///    and how they are to be handled.


    struct VariableInfo                         ///    Data kept for each "display" variable
    {

        VariableInfo(bool floating=false)
        :   inScope_(true)
        ,   floating_(floating)
        ,   chgCnt_(0)
        {
        }

        bool  awaitChange(size_t startValue, int maxDelay=5);
                                                ///<   Suspend the caller until maxDelay has been exceeded
                                                ///    or member chgCnt_ has changed.  
                                                ///    
                                                ///    Only chgCnt_ is protected by the mutex in this
                                                ///    class.
                                                ///
                                                ///    @param  startValue   Start value of chgCnt_
                                                ///    @param  maxDelay     Max time in seconds to wait
                                                ///                         before timeout.
                                                ///
                                                ///    @return true if a timeout has occurred.

        size_t changeCount()
                                                ///    Get chgCnt_ in a thread safe manner
        {
            ScopedMutexLock lock(mutex_);
            return chgCnt_;
        }

        void incrementChangeCount()
        {
            ScopedMutexLock lock(mutex_);
            ++chgCnt_;
            
        }


        bool        inScope_;                   ///<   Is this variable currently in scope?


        bool        floating_;                  ///<   A flag indicating that this variable is
                                                ///    supposed to be evaluated (by name) in the current
                                                ///    context each time gdb stops.
                                                ///
                                                ///    Floating variables should be interpreted
                                                ///    as patterns rather than variable references.
                                                ///
                                                ///    Consider the variable name "i".  It might 
                                                ///    represent many different actual variables
                                                ///    in many different contexts simultaneously.
                                                ///
                                                ///    When a variable is floating, gdb decides
                                                ///    which "i" is right for the current context.
                                                ///    This is helpful for recursive routines
                                                ///    wherein you might always want to display
                                                ///    a function parameter everytime you stop
                                                ///    in a recursive sequence.
                                                ///
                                                ///    Non floating variables always refer to the
                                                ///    same memory location where the variable
                                                ///    was first defined.

        std::string id_;                        ///<   The name GDB interpreter=mi mode variablename.
                                                ///    This name means nothing to the user but
                                                ///    is the string used to refer the variable
                                                ///    when using the -var-evaluate-expression
                                                ///    command to gdb.  For structures, this
                                                ///    has the following basic syntax:
                                                ///
                                                ///       <varname>.public.member1
                                                ///       <varname>.private.member2.public.m3
                                                ///       etc

        std::string expr_;                      ///<   The expression that gets @b conceptually gets
                                                ///    evaluated to obtain the value for this 
                                                ///    display variable.
                                                /// 
                                                ///    In practice, this not actually used but it 
                                                ///    is helpful for displaying to the user
                                                ///     

        std::string value_;                     ///<   The most recent value queried from gdb
                                                ///    for this expression.

        std::string type_;                      ///<   The "language" specific type name for this
                                                ///    variable.  Note that this value can
                                                ///    change if the variable is a "floating"
                                                ///    variable.

        std::map<std::string, VariableInfo> 
                                children_;      ///<   Child nodes

        void log(std::string const &align)const;///<   Given a string blanks to align the output, print the
                                                ///    variable and its substructures.
    private:                                           

        Mutex    mutex_;                        ///<   Protects only chgCnt_;
        size_t   chgCnt_;                       ///<   count of receipts of value_ from gdb

    };

    typedef std::map<std::string, VariableInfo>
            VariableValues;                     ///<   Data type representing the mapping between display variable
                                                ///    keys and their most recent values.

    typedef std::list<std::string>
            PendingVariables;                   ///<   Queue of display variable names to be populated from
                                                ///    from ^done,value= lines coming from gdb.  The assumption is
                                                ///    that a sequence of -var-evaluate-expression commands will
                                                ///    be sent and that each expression's name will be push_back()
                                                ///    onto this queue.  As ^done,value= lines come in, a pop_front()
                                                ///    will be done in order map names to values

    typedef std::list<std::string> IdList;      ///<   A list of variable or breakpoint ids.

    typedef std::map<std::string,
                     VarObj*> Breakpoints;      ///<   A mapping between breakpoint id string and the breakpoint contents.

    typedef std::map<std::string,
                     VarObj*> Threads;          ///<   A mapping between the thread id string and the thread contents.


    GdbController(bool diaglog, std::string const &logfileName);           
                                                ///<   Construct a controller but don't start the gdb or the target executable.
                                                ///    @param diaglog is a flag indicating that diagnostic logging
                                                ///           should be done.  This is true by default for testing
                                                ///           and development purposes, but applications should set it
                                                ///           to false;
                                                ///    @param logfileName is the name of a log file that will be 
                                                ///           written that contains the commands sent to the gdb debugger
                                                ///           and their respective outputs. 
                                                ///
                                                ///    Note that parameter, diaglog, controlls the volume of logging,
                                                ///    not its existence.

   ~GdbController();                            ///<   Stop all processes and go away


    ChildState state();                         ///<   @return the current state of the debugged child

    bool       requestSignalHandlerList();      ///<   Send the command to the debugger to provide the list of
                                                ///    signal handlers.  This function will sleep 1 second waiting
                                                ///    for the data to arrive.  If none does, it will return true
                                                ///    to indicate error.  It will also return true if the current
                                                ///    state of the debugger is not one of:  stopped, noStarted.
                                                ///
                                                ///    If the function returns false, all is well.  Call getSignalHandlers()
                                                ///    to obtain the list.

    bool       kill();                          ///<   Kill the debugged child if it is running.  The state
                                                ///    should either be unstartablke, terminated or notStarted after
                                                ///    this function has been called.
                                                ///
                                                ///    @return true if successful, else false

    bool       halt();                          ///<   Send SIGINT to the debugged child process and thus
                                                ///    bring it to a halt.  The state should either
                                                ///    be unstartable, notStarted, stopped, or terminated when this
                                                ///    function is complete.
                                                ///
                                                ///    @note:  this function only works correctly after
                                                ///    the child process has stopped once on its own
                                                ///    or has been manually (and painfully) stopped by
                                                ///    using the "kill -SIGINT" command on it.  Normally
                                                ///    a program stops at main and thus any time after
                                                ///    main has been reached, the halt command will work
                                                ///    as expected.  But, prior to main, as when debugging
                                                ///    static initializers, you may have to set breakpoint
                                                ///    at some helpful place before running the program.
                                                ///    @b Or, as mentioned before, use the kill command to
                                                ///    force the program to breakpoint.  Again, this is a
                                                ///    highly rare situation.
                                                ///
                                                ///    @return true if successfull.

    bool      run(int pid=0);                   ///<   Start or re-start the debugged child using the current
                                                ///    argument list.  The state should be terminated or
                                                ///    running after this function completes.  Note that if
                                                ///    the current state() is unstartable when run() is invoked,
                                                ///    it will return immediately and the state will remain
                                                ///    unstartable.
                                                ///
                                                ///    @note:  this function does not wait for execution to
                                                ///    complete but returns within 5 seconds in all cases.
                                                ///
                                                ///    @return true if successful.
                                                ///
                                                ///    @param  pid   An optional parameter which specifies the 
                                                ///                  process id of the running process to attach to.
                                                ///                  The parameter defaults to 0 which means "start a new 
                                                ///                  process" rather than attaching to an existing process.

    bool      resume();                         ///<   Continue the debugged child process' execution -- if the current
                                                ///    state is not unstartable.  If the child
                                                ///    is currently running, this call does nothing.
                                                ///    if the child is in any other state, it either
                                                ///    continues from the point of a halt or starts the
                                                ///    process over from scratch.
                                                ///
                                                ///    @note:  this function does not wait for execution to
                                                ///    complete but returns immediately under all circumstances.
                                                ///

    bool      stepOver();                       ///<   Continue the debugged child process' execution -- if the
                                                ///    state is not unstartable.  Only execute a single line of
                                                ///    code and do not descend into called functions.
                                                ///
                                                ///    @note:  this function does not wait for execution to
                                                ///    complete but returns immediately under all circumstances.
                                                ///
                                                ///    @return true if successful.

    bool      stepInto();                       ///<   Continue the debugged child process' execution -- if the
                                                ///    state is not unstartable.  Only execute a single line of
                                                ///    code but @b do descend into called functions.
                                                ///
                                                ///    @note:  this function does not wait for execution to
                                                ///    complete but returns immediately under all circumstances.
                                                ///
                                                ///    @return true if successful.

    bool stepOut();                             ///<   Run til the return of the current function;


    bool setArgs(std::string const &args);      ///<   Set the command line arguments to be used the next
                                                ///    time that the debugged program is restarted.

    bool setExecutable(std::string const &e);   ///<   Set the name of the executable to run as the debugged child next.
                                                ///    If the process is currently running or stopped,
                                                ///    kill it.
                                                ///
                                                ///    @return true if successful and the executable
                                                ///                 really exists, etc.

    void watchLocation(std::string const &frame, std::string const &expression, char const *access="");
                                                ///<   Add a breakpoint of type "watch point".  A watchpoint causes the
                                                ///    program to stop when an address is accessed using the specified
                                                ///    access type:
                                                ///
                                                ///      @li any access
                                                ///      @li any read
                                                ///      @li any write
                                                ///
                                                ///    @param  frame       The stack frame number, as a string.
                                                ///    @param  expression  The expression to check for access.
                                                ///    @param  access      An optional argument.  By default, any 
                                                ///                        write to the object referenced by the
                                                ///                        exception will trigger a stop.  However,
                                                ///                        Access can be specified as "-a" or "-r",
                                                ///                        meanying "any access" and "any read"
                                                ///                        respectively.
                                                ///
                                                ///    The expression should result in a reference to a variable.  Thus,
                                                ///    the expression can be "varname" or "*pointer", etc.

    void exceptionBreakpoints(bool catchPoint, bool throwPoint);  
                                                ///<   Add exception handling breakpoints.  
                                                ///    @param catchPoint  If true, add a breakpoint on the exception catch
                                                ///                       handling logic.
                                                ///    @param throwPoint  If true, add a breakpoint on the point of
                                                ///                       of throwing exceptions (all of them).

    void setBreakpointCondition(std::string const &bpID, std::string const &bpCond);
                                                ///<  Modify the condition associate with the specified breakpoint id.
                                                ///   @param  bpID   The breakpoint number to be modified.
                                                ///   @param  bpCond The expression which determines if the breakpoint
                                                ///                  should cause the debugger to stop or not.

    bool getBreakpointCondition(std::string const &bpID, std::string &bpCond);
                                                ///<  Get the condition associate with the specified breakpoint id.
                                                ///   @param  bpID   The breakpoint number to be modified.
                                                ///   @param  bpCond A reference to a string that will hold the returned
                                                ///                  breakpoint condition.


    std::string
    addBreakpoint(std::string const &file,
                            size_t        line);  ///<   Set a new breakpoint at a particular source file
                                                  ///    and line number.  Note that duplicates are ignored.
                                                  ///    @return breakpoint id string, if successful.  Empty
                                                  ///    string otherwise.
                                                  ///
                                                  ///    @param file  The name of the file where the breakpoint
                                                  ///                 should be set.
                                                  ///
                                                  ///    @param line  The number of the line where the breakpoing
                                                  ///                 should be set within the file
                                                  ///
                                                  ///    @return empty string for success, or an error message
                                                  ///
                                                  ///    @note:  If the process is not running, then the
                                                  ///    breakpoint will take affect at the next run() or
                                                  ///    resume() call.  If it is running, then the process will
                                                  ///    be temporarily stopped long enough to enable this
                                                  ///    breakpoint before it is restarted.
                                                  ///
                                                  ///    Note that the automatic start and restart may not
                                                  ///    good for the debug session (assuming the program
                                                  ///    is not in a stopped state).  Be advised.

    std::string
    addBreakpoint(std::string const &func);     ///<   Add a breakpoint on a named function or at an
                                                ///    address if you specify a hex number.
                                                ///
                                                ///    @param func The name of the function or address
                                                ///                at which to stop.
                                                ///
                                                ///    @return empty string for success or an error message

    std::string
    removeBreakpoint(std::string const &n);     ///<   Remove an existing breakpoint given its id string.
                                                ///    The id string can be obtained from the breakpoints
                                                ///    function's returned list.
                                                ///
                                                ///    @param n  The name of the breakpoint, as found in the
                                                ///              breakpoints() function return list.
                                                ///
                                                ///    @return empty string for success or a breakpoint number.

    std::string getMostRecentBreakpoint();      ///<   get the name of the most recently created breakpoint


    void enableBreakpoint(std::string const &name, bool enabled=true);  
                                                ///<  Set the enabled flag to be true or false for a specified
                                                ///   breakpoint whose name is supplied.
                                                ///
                                                ///   @param name   The name of the breakpoint as returned in the
                                                ///                 breakpoints() function return list.
                                                ///
                                                ///   @param enabled  If true, enable the breakpoint, else
                                                ///                   disable it.
    
    void removeMonitoredVariable(std::string 
                                     const &e); ///<   Delete the specified monitored variable given its name.
                                                ///    Names are obtained when you call addMonitoredVariable.
                                                ///
                                                ///    Note that the variables_ array is keyed by the display
                                                ///    variable name.  The name is not the same as the
                                                ///    expression that gets evaluated.


    std::string
    addMonitoredVariable(std::string const &e); ///<   Add an variable name expression to the monitored variable set.
                                                ///        @param e The expression to be evaluated.
                                                ///        @return  A string id for the new display expression
                                                ///                 or empty string if the display variable could
                                                ///                 not be set.

    void
    changeExpression(std::string const &e,
                     std::string const &value); ///<   Change the value of an existing expression.
                                                /// 
                                                ///    @param e      The expression to modify
                                                ///    @param value  The new value to assign to the expression
                                                ///
                                                ///    Note that "expressions" are "display variables".  You
                                                ///    can only change value of certain kinds of expressions and
                                                ///    if you try to change the wrong kind, you will get an error.


    TargetOutput &targetOutput()                ///    @return a reference to the queue
                                                ///    holding the output from child process
                                                ///    being debugged.
                                                ///
                                                ///    Note that the returned reference refers to an object
                                                ///    that contains a mutex of its own, so you don't need
                                                ///    worry about locking the GdbController object as a whole.
    {
        return targetOutput_;
    }


    std::string stateName()                     ///    Return the name of the current state as a string
    {
        return stateName(state());
    }

    static std::string stateName(ChildState);   ///<    Return the name of the specified state as a string

    void sendToTarget(std::string const &s);    ///     Write to the child's pty masterFD so the child can
                                                ///     read the data.

    void setState(ChildState, 
                  std::string const &);         ///<    Set the current state of the GdbController and
                                                ///     and a reason description for entering the state.
                                                ///     NOTE:  mutex locks *this.


    void printAndDeleteTargetOutput();          ///<    A debug tool that allows us to print and
                                                ///     discard the data previously detected from the
                                                ///     debugged child target process.

    std::string reason() const;                 ///<    Get the reason description for the most recent change
                                                ///     change.

    void adoptStopInfo(VarObj* reason);         ///<    Adopt a new pointer which points to an explanation of the most 
                                                ///     recent reason that gdb stopped.  The stop info is a composite VarObj 
                                                ///     with the following members:
                                                ///       @li reason
                                                ///       @li disp
                                                ///       @li bkptno
                                                ///       @li frame
                                                ///       @li thread-id
                                                ///       @li stopped-threads
                                                ///       @li core
                                                ///     The frame is also a composite VarObj with the following members
                                                ///       @li addr
                                                ///       @li func
                                                ///       @li args
                                                ///       @li file
                                                ///       @li fullname
                                                ///       @li line
                                                ///     The args is an Array object.  Each member has the following fields:
                                                ///       @li name
                                                ///       @li value
                                                ///
                                                ///     Within the frame object, the fullname will only exist if the filename
                                                ///     can actually be found.  The file value may be the only one available if
                                                ///     the source for the code found at a stop point is not available.

    void clearStopInfo();                       ///<    Remove the most recent stop info.

    std::string stopInfo() const;               ///<    Get the most recent stop info as a string -- parsable by VarObj.  May
                                                ///     be empty.

    void sendCommand(std::string const &c);     ///<    Send a command to gdb.  Assumes someone else has
                                                ///     a mutex lock.  Note that this like, um, seriously dangerous to
                                                ///     use!  Reserve it for debugging programs using GdbController.
                                                ///     Don't use it in a real application.


    void threads(IdList &) const;               ///<    Get a list of the ids of the currently active threads.

    std::string thread(std::string const 
                             &threadId) const;  ///<    Given a thread id, get the VarObj encoded data associated with it.


    VarObj const* changelist() const;           ///<    The list of variables which changed values during the last stop

    std::string   stack() const;                ///<    Get the stack frames associated with the most recent stop as a string.
                                                ///     Parsable by VarObj.  Could be empty.  The contents of the encoded string
                                                ///     is an array fro frame objects.  See ArrayVarObj.   Each member of the 
                                                ///     array is a "frame" object.  The frame objects have the following
                                                ///     components:
                                                ///
                                                ///         @li address
                                                ///         @li level
                                                ///         @li file  
                                                ///         @li fullname
                                                ///         @li func
                                                ///         @li args  -- seems to be broken

    std::string const &curThread() const;       ///<    Current thread at the time of the last stop.

    void variables(IdList &) const;             ///<    Get a list of the ids referencing display variables.  A display variable
                                                ///     is a symbolic name for an expression.  Expressions have complex attributes.
                                                ///     This function just gives you the ids of the existing variables.
                                                ///
                                                ///     @see variable() for information about a specific variable.

   VariableInfo const * variable(std::string const &
                        id) const;              ///<    Get information about a specific variable.  
                                                ///
                                                ///     @param id   The id of the desired variable.  @see @ref variables().
                                                ///
                                                ///     @return     The most recent value assigned to this display variable.
                                                ///
                                                ///     

    void breakpoints(IdList &ids) const;        ///<    Get a list of the ids of all existing breakpoints

    std::string breakpoint(std::string const&
                                    id) const;  ///<     Return a var-obj encoded version of the breakpoint info for
                                                ///      the specified breakpoint -- or empty string if the id is
                                                ///      not valid;
                                                ///
                                                ///      @param id   A string containg the id of a VarObj encoding of
                                                ///                  of the info for the specified breakpoint.
                                                ///
                                                ///      To use the return value, parse it and look at the fields.
                                                ///      You should get a composite var object with the following
                                                ///      members:
                                                ///
                                                ///        @li  number
                                                ///        @li  type
                                                ///        @li  disp
                                                ///        @li  enabled
                                                ///        @li  addr
                                                ///        @li  func
                                                ///        @li  file
                                                ///        @li  fullname
                                                ///        @li  line
                                                ///        @li  times
                                                ///        @li  original-location
                                                ///        

    void updateMonitoredVariableAttributes(
                      std::string const &name,
                      std::string const &expr,
                      std::string const &type
                                          );    ///<    Given the fully qualified name of a monitored variable name expression,
                                                ///     find the appropriate substructure that contains
                                                ///     it, or create said structure if needed, and update
                                                ///     its fields accordingly.
                                                ///     

    void logMonitoredVariables() const;         ///<    Describe all the monitored variables using CXXTLS_LOG.

    VariableInfo*
    findOrCreateMonitoredVariable(
                     std::string const &gdId 
                                         );     ///<    Given a gdb monitored variable object's "variable name expression"
                                                ///     Find or create the corresponding variables_ info.


    void triggerMonitoredVariableEvaluation(
                      std::string const &expr); ///<    Trigger gdb to re-evaluate a given expression.
                                                ///     The gdb output will be automatically picked up
                                                ///     by interpreterGdbOutput().
                                                ///
                                                ///     @note the expression has to be valid for
                                                ///     use in the gdb command:  "-var-expression-evaluate".
                                                ///     This means that expression must be structured
                                                ///     like this:  var.member.member...
                                                ///
                                                ///     Where "var" is one of the names returned by
                                                ///     the function, variables().  Further, the member
                                                ///     names must correspond to the gdb variable object
                                                ///     expression syntax.  Typically, they have to look
                                                ///     like this:
                                                ///@code
                                                ///
                                                ///     D12.public.member.private.member ...
                                                ///
                                                ///@endcode     
                                                ///     

    bool writeTargetStdin(std::string const &s);///<    Write string s to the debugged child process'
                                                ///     stdin stream.
                                                ///
                                                ///     @param s     The string to write
                                                ///     @return true if successfull.
                                                ///
                                                ///     Note that the input stream must have a trailing
                                                ///     newline in it if a newline is desired.  This
                                                ///     function will not add a newline automatically.

    bool awaitStateChange(size_t startChangeCount, int seconds=5);       
                                                ///<    Wait up to a specified number of seconds for the
                                                ///     state to change.  Note that the new state value, state_,
                                                ///     may still be the same as the old state value.  This
                                                ///     function waits until a call to setState() is made,
                                                ///     regardless of the value being set.  
                                                ///
                                                ///     @param startChangeCount  The current change count
                                                ///                              at the time the function is
                                                ///                              called (more or less).
                                                ///
                                                ///     @param seconds Optional count in seconds to wait.
                                                ///                    The default is 5 seconds.
                                                ///
                                                ///     @return true to indicate that a time out occurred.
                                                ///
                                                ///     Thread Safe!

    bool awaitStackParmChange(size_t startChangeCount, int seconds=5);       
                                                ///<    Wait up to a specified number of seconds for the
                                                ///     stack parm change count to change.  Note that the new stack parm
                                                ///     encoded strings if found in variable stkPrm_.
                                                ///
                                                ///     @param startChangeCount  The current change count
                                                ///                              at the time the function is
                                                ///                              called (more or less).
                                                ///
                                                ///     @param seconds Optional count in seconds to wait.
                                                ///                    The default is 5 seconds.
                                                ///
                                                ///     @return true to indicate that a time out occurred.
                                                ///
                                                ///     Thread Safe!

    bool awaitVariableChange(size_t startChangeCount, int seconds=5);       
                                                ///<    Wait up to a specified number of seconds for the
                                                ///     stack variable change count to change.  Note that the new variable
                                                ///     encoded strings if found in variable stkVar_.
                                                ///
                                                ///     @param startChangeCount  The current change count
                                                ///                              at the time the function is
                                                ///                              called (more or less).
                                                ///
                                                ///     @param seconds Optional count in seconds to wait.
                                                ///                    The default is 5 seconds.
                                                ///
                                                ///     @return true to indicate that a time out occurred.
                                                ///
                                                ///     Thread Safe!

    std::string stackFrameParms(std::string const &frameNumber);
                                                ///<    Get the function parameters associated with a 
                                                ///     specific stack frame.  The string returned is VarObj
                                                ///     encoded and must be parsed.  An empty or invalid
                                                ///     result string is possible.
                                                ///
                                                ///     @param frameNumber  A string containing a number.
                                                ///
                                                ///     @return  An ArrayVarObj encoded string containing
                                                ///              multiple CompositeVarObj encoded substrings
                                                ///              with the following members:
                                                ///              @li name
                                                ///              @li value
                                                ///     

    std::string stackFrameVariables(std::string const &frameNumber);
                                                ///<    Get the variables defined in a 
                                                ///     specific stack frame.  The string returned is VarObj
                                                ///     encoded and must be parsed.  An empty or invalid
                                                ///     result string is possible.
                                                ///
                                                ///     @param frameNumber  A string containing a number.
                                                ///
                                                ///     @return  An ArrayVarObj encoded string containing
                                                ///              multiple CompositeVarObj encoded substrings
                                                ///              with the following members:
                                                ///              @li name
                                                ///              @li value
                                                ///     

    std::string evaluateTmpExpression(std::string const &frame, std::string const &exp);
                                                ///<    Create a temporary expression, evaluate it, and return its value.
                                                ///
                                                ///     @param frame The stack frame in which the expression should be executed
                                                ///     @param exp   The expression to be evaluated.
                                                ///
                                                ///     @return The expression's value.  If the string is empty, then
                                                ///             either the expression was invalid or gdb was running --
                                                ///             can't evaluate expressions during this period.

   void fetchDisplayExpressionResults(std::vector<std::string> &results);
                                                ///<    In a thread safe manner, extract and remove the display expression
                                                ///     results from the most recent run of the target.  The display
                                                ///     expressions appear in a vector of strings.  Each string begins
                                                ///     with the following:
                                                ///
                                                ///     @li  integer : value
                                                ///
                                                ///     Where the integer is the number associated with the expression
                                                ///     variable and value is an encoded string describing the simple
                                                ///     object or class object that is represented by the expression.
                                                ///
                                                ///     @param[out]  results  The vector of strings displayed in this
                                                ///                           particular run of the executable.
                                                ///
                                                ///     Note that the displayed expressions are dependent on the scope
                                                ///     that the program count is sitting in.  The integer value 
                                                ///     identifying the expression is independent of the scope and is
                                                ///     thus unique across all scopes.

    void fetchDisplayExpressions(std::map<std::string,  std::string> &expressions);
                                                ///<    In a thread safe manner, extract the mapping between the
                                                ///     expression numbers and the expression text.  This mapping
                                                ///     does not describe the scope of the expression in any way.
                                                ///     The user just has to keep track of this in his head.

   bool displayExpressionResults();             ///<    @return true if there are display expression results

   bool createDisplayExpression(std::string const &frame, std::string const &expression);
                                                ///<    In the specified frame, add the specified expression
                                                ///     for display each time that the program stops within
                                                ///     the stack frame defined by the frame.  The stack frame
                                                ///     is a string representation of a number that corresponds
                                                ///     to the stack depth during the current stop stack frame.
                                                ///
                                                ///     @return true to indicate failure or timeout

   bool deleteDisplayExpression(std::string const &expression);
                                                ///<    Delete the specified automatic display expression.
                                                ///
                                                ///     The expression name is an integer that will have been
                                                ///     provided during the call to createDisplayExpression.
                                                ///     The integer is not the expression passed to that function,
                                                ///     rather it will be associated with the created expression.
                                                ///     See fetchDisplayExpressions() which gives you the names
                                                ///     (ie numbers) of the display expressions.
                                                ///
                                                ///     Display expression names are unique across stack frames.
                                                ///
                                                ///     @return true to indicate failure or timeout


    bool hasSymbols() const { return hasSymbols_; }     
                                                ///<  @return true if there are probably debugging symbols available


    void getSignalHandlers(SignalHandlers &output); 
                                                ///<    In a thread safe manner, copy out the
                                                ///     signal handling table

    std::string getcwd();                       ///<    get the current working directory
                                                ///     @return A string.  If empty, then an error occurred,
                                                ///             otherwise, the current working directory of the
                                                ///             debugger starget at the time of the start of
                                                ///             debugging.
    bool setcwd(std::string const &d);          ///<    Set the current working directory before running the
                                                ///     the child target process.
                                                ///     @param d  The name of the directory to set as
                                                ///               the current working directory.
                                                ///     @return   true if a failure occurred.

    bool handleSignal(std::string const &name,
                      std::string const &stop,
                      std::string const &print,
                      std::string const &pass
                     );                         ///<    Set the handling for a signal whose name is specified.
                                                ///     @param name  SIG<XXXX>  See gdb manual
                                                ///     @param stop  A flag meaning that gdb should stop when
                                                ///                  this signal arrives.  The flag is a string
                                                ///                  with either Y or N in it.
                                                ///     @param pass  A flag meaning that gdb should pass the 
                                                ///                  signal to the target.  The flag is a string
                                                ///                  with either Y or N in it.
                                                ///     
                                                ///     @param print A flag meaning that gdb should print a message
                                                ///                  when the signal occurs.  The flag is a string
                                                ///                  with either Y or N in it.


    std::string whatis(std::string const &frame, std::string const &expr);  
                                                  ///<  Ask the debugger what data type an expression is within a given stack frame.  
                                                  ///   @return empty string for error.  Otherwise, the type.

    void setSaveFile(std::string const &filename)
        /// The name of the file to read/write breakpoints at startup/shutdown.
    {
       saveFile_ = filename;
    }


    std::string saveFile() const
        ///  @return the name of the save file for this executable.
    {
        if(exe_.empty()  && saveFile_.empty() )
        {
            return "";
        }

        if(!saveFile_.empty())
           return saveFile_;

        return exe_ + std::string(".gdbMode");
    }

    bool saveState();                           ///<    Attempt to save state to the saveFile().  return true if error.

    bool selectStackFrame(std::string id);      ///<    Attempt to set the current stack frame location to id.  Return true if error

private:

    size_t stateChangeCount();                  ///<    return the number of times that the setState() function
                                                ///     has been called.  
                                                ///
                                                ///     Thread Safe!

    size_t argsChangeCount();                   ///<    return the number of times that the stkCnt_ variable
                                                ///     has been incremented.
                                                ///
                                                ///     Thread Safe!

    size_t variablesChangeCount();              ///<    return the number of times that the varCnt_ variable
                                                ///     has been incremented.
                                                ///
                                                ///     Thread Safe!

    void interpretGdbOutput(char const *line);  ///<    Interpret the meaning of a line of output from
                                                ///     target process being debugged.
                                                ///
                                                ///     @param line  A line of text produced ONLY by the
                                                ///                  GDB program itself to indicate its
                                                ///                  own state or that of a debugged
                                                ///                  child process.  Child program output
                                                ///                  should not be sent to this function.
                                                ///                  This separation is accomplished using
                                                ///                  a pseudo tty to run the target process
                                                ///                  in.  See "gdb -tty /dev/pts/XX".
                                                ///
                                                ///     Note that this function manipulates members of
                                                ///     this class and thus mutex locks the class for
                                                ///     brief periods.

    bool startGdb();                            ///<    Start the gdb_ controlled process and also the
                                                ///     thread that will monitor it.  Note that this
                                                ///     does not start the child being debugged -- or even
                                                ///     set it's name.  When successful, the state() member
                                                ///     will return "notStarted".  If the function fails,
                                                ///     the state() function will return unstartable.  This
                                                ///     means that the run() command will not work.

    void stopGdb();                             ///<    Kills the debugged child, if any.  Terminates
                                                ///     gdb, waits for the slaveDriver to join,
                                                ///     then returns.

    void sendArgsAndSetup();                    ///<    Send exe name, program arguments, and other initialization
                                                ///     to gdb.

    std::string nextId(char type);              ///<    Get a new unique id given a single character which defines
                                                ///     the type of id you are getting.  Ids are of the form:  "T123"
                                                ///     where T is the type and 123 is a number that makes the identifier
                                                ///     that makes the return value unique across all types.
                                                ///
                                                ///     @param type  A single charactering defining the type of id.
                                                ///     @return      A unique string.


    TargetOutput targetOutput_;                 ///<    The debugged target process's output.  Note that
                                                ///     this queue is guarded by mutex of its own.

    std::auto_ptr< ControlledProcess > gdb_;    ///<    Gdb running in controlled mode.

    std::auto_ptr< PseudoTerminal    > pty_;    ///<    A pty for gdb to run the target in.

    ThreadId                           gdbRdr_; ///<    The thread that reads from gdb's interpreter=mi
                                                ///     output stream.

    ThreadId                           tgtRdr_; ///<    The thread that reads the running target's
                                                ///     output.

    pid_t                                pid_;  ///<    Debugged Target process id
    ChildState                         state_;  ///<    Current state of the debugged child, if any.
    std::string                        reason_; ///<    The reason for the most recent state change.
    size_t                             chgCnt_; ///<    Number of calls to setState
    size_t                             stkCnt_; ///<    stack parm change count
    size_t                             varCnt_; ///<    Count changes to stkVar_;
    std::string                        stkPrm_; ///<    The most recently requested stack frame function arguments
    std::string                        stkVar_; ///<    variables for this stack frame
    std::string                        expVal_; ///<    The value= string from a -data-evaluate-expression statement.
    size_t                             expCnt_; ///<    count of changes to expVal_
    bool                               expWat_; ///<    currently waiting for a display expression value
    Mutex                              stkMut_; ///<    The mutex guarding the stk arguments info
    
    Mutex                              mutex_;  ///<    Mutex guard for this whole class object.
                                                ///     Note that InterpretGdbOutput locks this as needed.

    std::string                        args_;   ///<    The run arguments for the debugged child target process.
    std::string                        exe_;    ///<    The name of the debugged child target executable.
    std::string                    cdForRun_;   ///<    when you call setcd before the first start, this gets
                                                ///     populated.  It is cleared by the run() process

    bool                               dirty_;  ///<    This flag means either args or exe name have changed


    static char const           *stateName_[];  ///<    Array of names corresponding to ChildStates

    std::list<std::string>             pndBP_;  ///<    As yet unasserted breakpoints.  Usually set prior
                                                ///     to the first run.

    std::auto_ptr<VarObj>           stopInfo_;  ///<    Information about the most recent stop detected by gdb

    std::auto_ptr<VarObj>            threads_;  ///<    Threads active at the most recent stop
    std::auto_ptr<VarObj>         changelist_;  ///<    Variables changed in the last stop
    std::string                    curThread_;  ///<    The current thread id
    VariableValues                 variables_;  ///<    Most recent state of display variable values
    PendingVariables                 pndVars_;  ///<    @see PendingVariables for details
    Breakpoints                  breakpoints_;  ///<    Map of id to breakpoint.
    std::string                 mrBreakpoint_;  ///<    Most recent breakpoint accessed (to assist the UI

    Mutex                     stopping_mutex_;  ///<    Used by stopGdb to communicate with the gdbConsoleTgtReader()
    bool                            stopping_;  ///<    Used by stopGdb to communicate with the gdbConsoleTgtReader()

    bool                        debugLogging_;  ///<    True by default, see quiet(), which silences this.

    std::auto_ptr<VarObj>             stack_;   ///<    stack frame at most recent stop point (up to 40 frames)


    std::auto_ptr<std::ostream>       gdbLog_;  ///<    A log of gdb interactions.  Commands sent to gdb are intermingled
                                                ///     without output from the debugged target executable.

    Mutex                             logMut_;  ///<    the mutex that guards log interactions

    char const *                      async_;   ///<    a substring used on some commands to enable them in asynchronous
                                                ///     operation.  Defaults to "&" but may become "" if this target
                                                ///     is not executed in the asynchronous operationg mode.

    Mutex                             eMutex_;  ///<    The mutex that guards enames and evalues
    bool                              eError_;  ///<    True if an error occurred while waiting for an expression to change
    std::map<std::string,std::string> enames_;  ///<    A mapping between expression integer (strings) and the 
                                                ///     program scope specific expressions that they map to.

    std::vector<std::string>         evalues_;  ///<    The display expression values.  They are formatted like this:
                                                ///     @li  integer : encodedValue
                                                ///     The encoded value can refer to a number, a string, or a structure.

    std::string                      eaccum_;  ///<     An accumulator to hold fragments of one evalue as it is built.
    size_t                           eChgCnt_; ///<     count of changes to evalues_
    bool                          hasSymbols_; ///<     True if symbols are probably available.  False
                                               ///      only if they are known not to exist.

    SignalHandlers            signalHandlers_; ///<     Current signal handling behavior table.

    size_t                         curDirCnt_; ///<     The count of changes to curdir_
    std::string                       curDir_; ///<     The string containing the result of a pwd command cwd

    Mutex                     doneErrorMutex_; ///<     Guards doneErrorCnt_
    size_t                      doneErrorCnt_; ///<     Count of the number of ^done or ^error records from gdb
    std::string                 doneErrorStr_; ///<     Most recent ^error,msg="string".
                                               ///      Should be empty most of the time.  It gets recent by
                                               ///      awaitDoneErrorChange() and previous value is returned.
    std::string                doneWhatisStr_; ///<     protected by doneErrorMutex_, will contain the the
                                               ///      gdb 'whatis command' result if one is encountered.
    bool                   doneErrorScooping_; ///<     When true, all gdb output of the form &"..." will be
                                               ///      appended to doneScooped_
    std::vector<std::string>     doneScooped_; ///<     Golly gee, look what I done scooped up!
    bool                        firstSetArgs_; ///<     first time set args is called
    std::string                     saveFile_; ///<     file where breakpoints are read/written 

public:

    static void incrementReaderThreadCount();   ///<     increment the count of reader threads.  There should be 0, 1, or 2.
                                                ///      0 means that the child is not running.  1 means we are in a start up or
                                                ///      shutdown state.  2 means that we are fully running normally.

    static void decrementReaderThreadCount();  ///<     decrement the count of reader threads.  There should be 0, 1, or 2.
                                               ///      0 means that the child is not running.  1 means we are in a start up or
                                               ///      shutdown state.  2 means that we are fully running normally.

    static int readerThreadCount();            ///<     @return 0, 1, or 2: the count of reader threads:
                                               ///      0 means that the child is not running.  1 means we are in a start up or
                                               ///      shutdown state.  2 means that we are fully running normally.


    bool doneErrorScoop(std::string const &command, 
                        std::vector<std::string> &output
                       );                      ///< Issue a gdb command and scoop up results.
                                               ///  @return true if error occurred
                                               ///  @param command  The gdb command to execute
                                               ///  @param[out]  output  The string vector to hold the 
                                               ///                       results.  If an error occurs,
                                               ///                       one output string will be  
                                               ///                       produced which describes the
                                               ///                       error.

   void writeGdbLogMessage(std::string const &s);

private:

    size_t doneErrorCount()
        ///  return the current doneErrorCnt_;  You should probably be calling clearDoneErrorMessage() instead.
    {
        ScopedMutexLock lock(doneErrorMutex_);
        return doneErrorCnt_;
    }

    size_t clearDoneErrorMessage()
        ///  Clear the doneErrorStr_ and doneWhatisStr_ variables and return the current value of
        ///  doneErrorCnt_.  Use this before issuing a sendCommand() if you are interested in
        ///  ^done, or ^error results lines for the command.
    {
        ScopedMutexLock lock(doneErrorMutex_);
        doneErrorStr_.resize(0);
        doneWhatisStr_.resize(0);
        return doneErrorCnt_;
       
    }

    void incDoneErrorCount(std::string const &errorString="")
        /// In a thread safe manner, increment the doneErrorCnt_ variable and set the
        /// error string.
    {
        ScopedMutexLock lock(doneErrorMutex_);
        ++doneErrorCnt_;

        doneErrorStr_ = errorString;
    }

    std::string awaitDoneErrorChange(size_t startCount, int seconds=5);
        ///< wait up to the specified number of seconds for the doneErrorCount()
        /// to change.  @return either:
        /// @li empty string to indicate success
        /// @li an error message string returned from gdb pas part of '^error,msg="string"' records
        /// @li "time out"


    void setSignalHandling(std::string const&);  ///<    Handle a gdb output line documenting a
                                                 ///     signal's handling logic.  The lines processed
                                                 ///     by this function look like this:
                                                 ///
                                                 ///     ~"SIG<NAME>   Yes\tNo\tYes\tDescription\n"
                                                 ///
                                                 ///     This function updates signalHandlers.

    bool awaitCwdChange(size_t startCount, int maxWait=5);  ///< await a change in curDirCnt, up to maxWait seconds.
    size_t getCwdChangeCnt() { return curDirCnt_; }         ///< return the count of changes to curDir_; 

    size_t  expressionChangeCount() 
    {
        ScopedMutexLock lock(stkMut_);
        return expCnt_;
    }

    std::string expressionValue() 
    {
        ScopedMutexLock lock(stkMut_);
        return expVal_;
    }

    size_t displayVariableChangeCount()
    {
       ScopedMutexLock lock(eMutex_);

       return eChgCnt_;
    }

    bool awaitExpressionChange(size_t startCount, int maxWait=5);
                                                ///<  Wait for the completion of a command:  -data-evaluate-expression.
                                                ///   
                                                ///   @param startCount   The start value for expCnt_, as obtained by expressionChangeCount().
                                                ///   @param maxWait      The maximum time in seconds to wait.
                                                ///
                                                ///   @return true to indicate timeout.
                                                ///
                                                ///   If no timeout, call expressionValue() to get the data

    bool awaitDisplayVariableChange(size_t startCount, int maxWait=5);
                                                ///<  Wait for the completion of a command:  'display expression'
                                                ///   
                                                ///   @param startCount   The start value for eChgCnt_, as obtained by expressionChangeCount().
                                                ///   @param maxWait      The maximum time in seconds to wait.
                                                ///
                                                ///   @return true to indicate timeout.
                                                ///
                                                ///   If no timeout, call expressionValue() to get the data



    void recordDisplayExpressionResults();      ///<  Add line to evalues_ if needed.
    void accumulateDisplayExpressionFragment(std::string const &line);
                                                ///<  Add a fragment to eaccum_ for later use by recordDisplayRexpressionResults

    void signalDisplayExpressionError(std::string const &msg);  ///< cancel any pending wait for a display expression value
                                                                ///  

public:
    Mutex controllerStartMutex_;  // used by the run() function to determine if we need to start the the controller thread.
    bool  controllerStarted_;     // only once for a given GdbController object's lifetime

private:

}; // class GdbController


}; // namespace cxxtls

#endif // not _MSCVER
#endif
