#ifndef __varobj_header_included_p
#define __varobj_header_included_p
//
// Copyright 2013, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

///@file
/// This file declares the class cxxtls::VarObj which is a representation
/// of a leave or composite node tree.
#include <string>
#include <map>
#include <vector>
namespace cxxtls
{

//===========================================================================
//===========================================================================
//===========================================================================

class VarObj
    ///  A VarObj is a representation of a (possibly) structured value.  VarObj
    ///  instances are easily parsable from normal text strings.  They are
    ///  can be any one of the following types:
    ///
    ///      @li a string  (leaf object)
    ///      @li a number  (leaf object)
    ///      @li a structure (composite object)
    ///      @li an array (non-leaf object)
    ///
    ///  Here are some example textual representations of VarObj:
    ///@code
    ///
    ///    47
    ///    "something\nNextLine"
    ///    {fred=13, bill={a=b, c=d}}
    ///    [1,2,3,"4",{name=value},tom="bill"]
    ///
    ///@endcode
    ///
    /// Note that when specifying a composite object, members of the 
    /// composite must be named. 
    ///
{
public:

    struct ParseResult
    {
        VarObj* rv_;      // You must delete this!
        std::string error_;  // empty is all is well

        ParseResult(std::string const &error)
        :  rv_(0)
        ,  error_(error)
        {
            /// construct a failed parse error code. 
        }

        ParseResult(VarObj* rv)
        :  rv_(rv)
        ,  error_("")
        {
           /// construct a successful parse result
        }
    };

    static ParseResult readAndParseFile(std::string const &filename);
        ///<  @return a ParseResult object created when the specified
        ///   filename is read and parsed.



    static ParseResult parse(char const *&begin, char const *&end);
        ///<  static method to parse VarObj from a string fragment.
        ///   Note that the following list defines successfully parsable
        ///   values:
        ///@code
        ///
        ///   1234 
        ///   -1234e+245
        ///
        ///   "stuff\nmorestuff"
        ///
        ///   {a=<value1>, b=<value2>...}
        ///
        ///@endcode

    static ParseResult parse(std::string const &s)
        ///  static method to parse a string
    {
       char const *f = s.data();
       char const *l = f + s.size();

       return parse(f, l);
    }

    static ParseResult parseList(char const *&begin, char const *&end);
        ///<   A static method to parse a comma separated list of
        ///    name=value pairs from a string fragment.  Where the name is a 
        ///    standard C langauge identifier and the value is a normal 
        ///    VarObj representation.
        ///
        ///    Note that the parsing terminates on end or if a '} is found.
        ///
        ///

    static ParseResult parseList(std::string const &s)
        ///  static method to parse a string
    {
       char const *f = s.data();
       char const *l = f + s.size();

       return parseList(f, l);
    }

    static std::string encodeString(std::string const &value); 
        ///<  @return a VarObj parsable string given a raw string value.  The
        ///   return value will begin and end with '"' and will have escape
        ///   sequences that can be parsed via VarObj.

    virtual std::string value() const=0;
        ///<   Convert *this into a undecorated string.  @return a string value.

    virtual std::string parsable_value() const;
        ///<   Convert *this into a parsable (decorated) string.  @return a string value.

    virtual ~VarObj(); 
        ///<   Virtual destructor.

    virtual VarObj*  findChild(std::string const &);
        ///<  Search this object for a child member by a specified name.
        ///   @return a pointer to the child VarObj whose name is
        ///   specified, or return 0 if no such member exists.

    VarObj* findNestedChild(std::string const &);
        ///<  Search for a child using a "/" separated list of parent nodes. 
        ///   For example:
        ///@code
        /// 
        ///       fred/bill/sue
        /// 
        ///@endcode
        ///
        /// @return nul if not found

    VarObj const* findNestedChild(std::string const &) const;
        ///<  Search for a child using a "/" separated list of parent nodes. 
        ///   For example:
        ///@code
        /// 
        ///       fred/bill/sue
        /// 
        ///@endcode
        ///
        /// @return nul if not found

    virtual bool leaf() const; 
        ///<   Return true if this is a leaf node in the VarObj structure
       
}; // class VarObj

//===========================================================================
//===========================================================================
//===========================================================================

class StringVarObj
    ///  A VarObj that represents a simple string.
:  public VarObj
{
    std::string value_;

public:

   StringVarObj(std::string const &s)
   :  value_(s)
   {
   }

    virtual std::string value() const;
        ///<   Convert *this into a parsable string.  @return a string value.

    virtual std::string parsable_value() const;
        ///<   Convert *this into a parsable (decorated) string.  @return a string value.

    
};
//===========================================================================
//===========================================================================
//===========================================================================
class IntegerVarObj
    ///  A VarObj that represents a simple integer.
:  public VarObj
{
    long value_;

public:

   IntegerVarObj(long s)
   :  value_(s)
   {
   }

    virtual std::string value() const;
        ///<   Convert *this into a parsable string.  @return a string value.
};

//===========================================================================
//===========================================================================
//===========================================================================
class FloatVarObj
    ///  A VarObj that represents a double.
:  public VarObj
{
    double value_;

public:

   FloatVarObj(double s)
   :  value_(s)
   {
   }

    virtual std::string value() const;
        ///<   Convert *this into a parsable string.  @return a string value.
};

//===========================================================================
//===========================================================================
//===========================================================================

class CompositeVarObj
    ///  A VarObj that represents a structure (ie composite object)
:  public VarObj
{
   typedef   std::map<std::string, VarObj*> ValueType;

   ValueType value_;

public:

   CompositeVarObj()
   {
   }

   ~CompositeVarObj();

    virtual bool leaf() const; 
        ///<   Return true if this is a leaf node in the VarObj structure

    VarObj*  findChild(std::string const &);
        ///<   Search this object for a child member by a specified name.
        ///   @return a pointer to the child VarObj whose name is
        ///   specified, or return 0 if no such member exists.

    VarObj const*  findChild(std::string const &child) const
        ///   Search this object for a child member by a specified name.
        ///   @return a pointer to the child VarObj whose name is
        ///   specified, or return 0 if no such member exists.
    {
       return (const_cast<CompositeVarObj*>(this))->findChild(child);
    }

    virtual std::string value() const;
        ///<   Convert *this into a parsable string.  @return a string value.

    VarObj*   release(std::string const &childName);
        ///<  Find and release the named child.  
        ///
        ///   @param childName  The name of the immediate child of this
        ///                     CompositeVarObj.  
        ///
        ///   @return NULL or a pointer to the extracted (and now detached)
        ///           child object.  You are now responsible for deleting
        ///           said item.


    void adopt(std::string const name, VarObj* value)
        ///<  Include a named value in the container and take responsibility
        ///   to delete said item.
        ///
        ///   @param name  The name of the element to add
        ///   @param value A pointer to a child var object which is to be
        ///                known by the name given.
    {
        value_.insert(  ValueType::value_type(name, value) );
    }

    bool empty() const
        ///<   return true if this object has NO children.
    {
       return value_.empty();
    }

};

//===========================================================================
//===========================================================================
//===========================================================================

class ArrayVarObj
    ///  A VarObj that represents an array of VarObj
:  public VarObj
{
   typedef   std::vector<VarObj*> ValueType;

   ValueType value_;

public:

   ArrayVarObj()
   {
   }

   ~ArrayVarObj();

    virtual bool leaf() const; 
        ///<   Return true if this is a leaf node in the VarObj structure

    virtual std::string value() const;
        ///<   Convert *this into a parsable string.  @return a string value.

    size_t  size() const
        ///<  Return the count of items in this container.
    {
       return value_.size();
    }

    VarObj* operator[](size_t i);
        ///<  Get a mutable pointer to the item at the specified location.  If the item does not exist,
        ///   then a null will be returned.
        ///
        ///   @param i  The index into the vector.
        ///   @return   Null or a pointer to item of interest.

    VarObj const* operator[](size_t i) const;
        ///<  Get a const pointer to the item at the specified location.  If the item does not exist,
        ///   then a null will be returned.
        ///
        ///   @param i  The index into the vector.
        ///   @return   Null or a pointer to item of interest.

    void adopt(VarObj* value)
        ///<  Add a new VarObj to the this vector and take responsibility to delete it.
    {
        value_.push_back(  ValueType::value_type(value) );
    }

    bool empty() const
        ///<   return true if this object has NO children.
    {
       return value_.empty();
    }

};



//===========================================================================
//===========================================================================
//===========================================================================


} // namespace cxxtls

#endif
